# © 2019 Noel Kaczmarek
import json


class Message:
    def __init__(self, author, data, type='message'):
        if type == 'message':
            self.data = json.dumps(data)
        elif type == 'bytes':
            self.data = data

        self.header = {'author': author, 'length': len(self.data), 'type': type}

    def GetHeader(self):
        return self.header

    def GetData(self):
        return self.data


# © 2019 Noel Kaczmarek
import os


class Shard:
    def __init__(self, filepath, offset, index, size, data):
        print('Creating shard...')
        self.file = filepath
        self.offset = offset
        self.index = index
        self.size = size
        self.data = data

    def Read(self):
        with open(self.file, 'rb') as file:
            file.seek(self.offset, 0)
            content = file.read(self.size)
            file.close()

        return content

    def Write(self):
        with open(self.file, 'wb') as file:
            file.seek(0, 0)
            file.write(self.data)
            file.close()

    def GetFile(self):
        return self.file

    def GetOffset(self):
        return self.offset

    def GetIndex(self):
        return self.index

    def GetSize(self):
        return self.size

    def GetData(self):
        return self.data


class File:
    def __init__(self, filepath):
        self.file = filepath
        self.size = os.stat(self.file).st_size
        self.shards = []
        print('File name: %s, File size: %d' % (self.file, self.size))

    def Read(self, bytes=0, offset=0):
        with open(self.file, 'rb') as file:
            file.seek(offset, 0)
            content = file.read(bytes)
            file.close()

        return content

    def Split(self):
        shard_size = round(self.size / 4)
        print('Shard size: %d' % shard_size)
        n = 0

        while n < 4:
            data = self.Read(shard_size, shard_size * n)
            self.shards.append(Shard('%d_%s.shard' % (n, self.file), shard_size * n, n, shard_size, data))
            n += 1

        for shard in self.shards:
            shard.Write()

    def Weld(self):
        print('Welding file together...')
        file = open('welded_' + self.file, 'wb')
        
        for shard in self.shards:
            file.seek(shard.offset, 0)
            file.write(shard.GetData())

        file.close()


# © 2019 Noel Kaczmarek
from P2P.connection import Connection
from P2P.message import Message

import threading
import socket
import json
import time
import uuid
import os


class Node(threading.Thread):
    def __init__(self, host, port):
        super(Node, self).__init__()

        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.terminate_flag = threading.Event()
        self.id = uuid.uuid4().int
        self.host = host
        self.port = port
        self.nodesIn = []
        self.nodesOut = []

        print('Host: %s, Port: %s, ID: %s' % (self.host, self.port, self.id))

        self.InitialiseServer()

    def InitialiseServer(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.bind(('', self.port))
        self.sock.settimeout(10.0)
        self.sock.listen(1)

    def Connect(self, host, port):
        if (host == self.host and port == self.port):
            return;

        for node in self.nodesOut:
            if (node.GetHost() == host and node.GetPort() == port):
                print('Connect: Already connected to this node.')
                return True

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            print('Connecting to %s port %s...' % (host, port))
            sock.connect((host, port))

            client_thread = Connection(self, sock, (host, port))
            client_thread.start()
            client_thread.Send({'type': 'InitialTransmission', 'ID': self.id, 'port': self.port})
            self.nodesOut.append(client_thread)
            self.PrintConnections()

        except Exception as e:
            print('TcpServer.connect_with_node: Could not connect with node. (' + str(e) + ')')

    def Disconnect(self, node):
        if node in self.nodesOut:
            node.Send({'type': 'TerminationSignal'}) # Update this line
            node.stop()
            node.join()
            del self.nodesOut[self.nodesOut.index(node)]

    def Leave(self):
        self.Broadcast({'type': 'LeaveRequest'})

    def Broadcast(self, data, exclude=[]):
        for node in self.nodesIn:
            if node not in exclude:
                self.SendToNode(node, data)

        for node in self.nodesOut:
            if node not in exclude:
                self.SendToNode(node, data)

    def SendToNode(self, node, data):
        if node in self.nodesIn or node in self.nodesOut:
            node.Send(data)
        else:
            print('SendToNode: Could not send the data, node is not found!')

    def SendRoutingTable(self, exclude=[]):
        nodesIn = []
        nodesOut = []

        for node in self.nodesIn:
            id = node.GetID()
            host = node.GetHost()
            port = node.GetPort()

            nodesIn.append({'ID': id, 'host': host, 'port': port})

        for node in self.nodesOut:
            id = node.GetID()
            host = node.GetHost()
            port = node.GetPort()

            nodesOut.append({'ID': id, 'host': host, 'port': port})

        self.Broadcast({'type': 'RoutingTableUpdate', 'IncomingConnections': nodesIn, 'OutgoingConnections': nodesOut}, exclude)

    def run(self):
        while not self.terminate_flag.is_set():
            try:
                connection, client_address = self.sock.accept()

                thread_client = Connection(self, connection, client_address)
                thread_client.start()
                self.nodesIn.append(thread_client)
                self.PrintConnections()

            except socket.timeout:
                pass

            except KeyboardInterrupt:
                os._exit(0)

            except:
                raise

            time.sleep(0.01)

        print('Node stopping...')

        for t in self.nodesIn:
            t.stop()

        for t in self.nodesOut:
            t.stop()

        time.sleep(1)

        for t in self.nodesIn:
            t.join()

        for t in self.nodesOut:
            t.join()

        self.sock.close()

        print('Node stopped')

    def stop(self):
        self.terminate_flag.set()

    def PrintConnections(self):
        print('Connections:')
        print('- Total incoming connections: %d' % len(self.nodesIn))
        print('- Total outgoing connections: %d' % len(self.nodesOut))

    def RemoveNodeFromTable(self, id):
        for node in self.nodeServer.nodesIn:
            if node['ID'] == id:
                self.nodeServer.nodesIn.remove(node)

        for node in self.nodeServer.nodesOut:
            if node['ID'] == id:
                self.nodeServer.nodesOut.remove(node)

    def IsConnectedTo(self, id):
        if id in self.nodesIn or self.nodesOut:
            return True
        else:
            return False

    def GetNodeInfo(self):
        return {'ID': self.id, 'port': self.port, 'IncomingConnections': len(self.nodesIn), 'OutgoingConnections': len(self.nodesOut)}

    def GetHost(self):
        return self.host

    def GetPort(self):
        return self.port

    def GetRoutingTable(self):
        return self.peers

    def GetID(self):
        return self.id


# © 2019 Noel Kaczmarek
from P2P.message import Message
from P2P.vars import HEADER_SIZE

import threading
import socket
import json


class Connection(threading.Thread):
    def __init__(self, nodeServer, sock, clientAddress):
        super(Connection, self).__init__()

        self.id = 0
        self.sock = sock
        self.host = clientAddress[0]
        self.port = clientAddress[1]
        self.nodeServer = nodeServer
        self.clientAddress = clientAddress
        self.terminate_flag = threading.Event()

    def Send(self, data):
        try:
            message = Message(self.nodeServer.GetID(), data)

            self.sock.sendall(json.dumps(message.GetHeader()).encode('utf-8'))
            self.sock.sendall(message.GetData().encode('utf-8'))
        except Exception as e:
            print('Connection.Send: Unexpected error:' + e)
            self.terminate_flag.set()

    def stop(self):
        self.terminate_flag.set()

    def run(self):
        while not self.terminate_flag.is_set():
            header = json.loads(self.sock.recv(HEADER_SIZE))
            data = self.sock.recv(int(header['length'])).decode('utf-8')

            # print('Header: %s' % header)
            # print('Data: %s' % data)

            if header['type'] == 'message':
                data = json.loads(data)

                if data['type'] == 'InitialTransmission':
                    self.id = data['ID']
                    self.port = data['port']
                    self.nodeServer.SendRoutingTable()

                    # print('ID: %s, Port: %d' % (self.id, self.port))

                elif data['type'] == 'LeaveRequest':
                    self.nodeServer.RemoveNodeFromTable(header['author'])

                elif data['type'] == 'RoutingTableUpdate':
                    for node in data['IncomingConnections']:
                        if node['ID'] != self.nodeServer.GetID():
                            self.nodeServer.Connect(node['host'], node['port'])

                    for node in data['OutgoingConnections']:
                        if node['ID'] != self.nodeServer.GetID():
                            self.nodeServer.Connect(node['host'], node['port'])

    def GetID(self):
        return self.id

    def GetHost(self):
        return self.host

    def GetPort(self):
        return self.port
